/* Parameters */
#ifndef SIZE_OF_TEST
#define SIZE_OF_TEST 100000
#endif
#ifndef NUMBER_OF_RUN
#define NUMBER_OF_RUN 20
#endif
#ifndef AVAIL
#define AVAIL 64
#endif
#define N 3
#define AFF_INCR 0
/* Includes */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>
#include <time.h>
#include <limits.h>
#include "utils.h"
#include "outs.h"
#include "affinity.h"

/* params */
static int verbose ;
static int size_of_test = SIZE_OF_TEST ;
static int speedcheck = 1 ;
static int stop_now = 0 ;
#define SYNC_K 128
#define SYNC_N 0

/* Full memory barrier */

inline static void mbar(void) {  asm __volatile__ ("sync" ::: "memory"); }

/* Barriers macros */

inline static void barrier_wait(unsigned int id, unsigned int k, int volatile *b) {
  if ((k % N) == id) {
    *b = 1 ;
  } else {
    while (*b == 0) ;
  }
}
/**********************/
/* Context definition */
/**********************/

typedef struct {
/* Shared variables */
  int* *y;
  int* *x;
  int* *ok;
  int *mem_y ;
  int *mem_x ;
  int *mem_ok ;
  int *_idx;
  int do_shuffle;
/* Final content of observed  registers */
  int *out_1_r1;
  int *out_2_r3;
/* Check data */
  pb_t *fst_barrier;
  po_t *s_or ;
  int* cpy_ok[N] ;
  int* cpy_y[N] ;
/* Barrier for litmus loop */
  int volatile *barrier;
  st_t seed;
/* Sizes */
  int _size_of_test;
  /* Logical proc -> core */
  cpus_t *cm;
} ctx_t;


/**********************/
/* Outcome collection */
/**********************/

#define NOUTS 4
typedef int outcome_t[NOUTS];

static const int out_1_r1_f = 0 ;
static const int out_2_r3_f = 1 ;
static const int ok_f = 2 ;
static const int y_f = 3 ;


typedef struct hist_t {
  outs_t *outcomes ;
  count_t n_pos,n_neg ;
} hist_t ;

static hist_t *alloc_hist(void) {
  hist_t *p = malloc_check(sizeof(*p)) ;
  p->outcomes = NULL ;
  p->n_pos = p->n_neg = 0 ;
  return p ;
}

static void free_hist(hist_t *h) {
  free_outs(h->outcomes) ;
  free(h) ;
}

static void add_outcome(hist_t *h, count_t v, outcome_t o) {
  h->outcomes = add_outcome_outs(h->outcomes,o,NOUTS,v) ;
}

static void merge_hists(hist_t *h0, hist_t *h1) {
  h0->n_pos += h1->n_pos ;
  h0->n_neg += h1->n_neg ;
  h0->outcomes = merge_outs(h0->outcomes,h1->outcomes,NOUTS) ;
}

static count_t sum_hist(hist_t *h) {
  return sum_outs(h->outcomes) ;
}

static void do_dump_outcome(FILE *fhist, int *o, count_t c) {
  fprintf(fhist,"%-6"PCTR":>1:r1=%i; 2:r3=%i; ok=%i; y=%i;\n",c,o[out_1_r1_f],o[out_2_r3_f],o[ok_f],o[y_f]);
}

static void just_dump_outcomes(FILE *fhist, hist_t *h) {
  outcome_t buff ;
  dump_outs(fhist,do_dump_outcome,h->outcomes,buff,NOUTS) ;
}

static int dump_hist(FILE *fhist,hist_t *h) {
  fprintf(fhist,"Test WRW+WR+lwsyncaa+syncpa Allowed\n") ;
  fprintf(fhist,"Histogram (%"PCTR" states)\n",finals_outs(h->outcomes)) ;
  just_dump_outcomes(fhist,h) ;
  int cond = h->n_pos > 0;
  fprintf(fhist,"%s\n",cond?"Ok":"No");
  fprintf(fhist,"\nWitnesses\n");
  fprintf(fhist,"Positive: %"PCTR", Negative: %"PCTR"\n",h->n_pos,h->n_neg);
  fprintf(fhist,"Condition exists (ok=1 /\\ y=2 /\\ 1:r1=1 /\\ 2:r3=0) is %svalidated\n",h->n_pos > 0 ? "" : "NOT ");
  return cond;
}

/*******************************************************/
/* Context allocation, freeing and reinitialization    */
/*******************************************************/

static void shuffle(ctx_t *_a) {
  if (rand_bit(&(_a->seed))) perm_ints(&(_a->seed),_a->_idx,size_of_test);
  for (int _i = _a->_size_of_test-1 ; _i >= 0 ; _i--) {
    _a->y[_i] = &(_a->mem_y[_a->_idx[_i]]);
  }
  if (rand_bit(&(_a->seed))) perm_ints(&(_a->seed),_a->_idx,size_of_test);
  for (int _i = _a->_size_of_test-1 ; _i >= 0 ; _i--) {
    _a->x[_i] = &(_a->mem_x[_a->_idx[_i]]);
  }
  if (rand_bit(&(_a->seed))) perm_ints(&(_a->seed),_a->_idx,size_of_test);
  for (int _i = _a->_size_of_test-1 ; _i >= 0 ; _i--) {
    _a->ok[_i] = &(_a->mem_ok[_a->_idx[_i]]);
  }
}

static void init(ctx_t *_a) {
  _a->_size_of_test = size_of_test;
  _a->seed = rand();
  _a->y = malloc_check(size_of_test*sizeof(*(_a->y)));
  _a->x = malloc_check(size_of_test*sizeof(*(_a->x)));
  _a->ok = malloc_check(size_of_test*sizeof(*(_a->ok)));
  _a->out_1_r1 = malloc_check(size_of_test*sizeof(*(_a->out_1_r1)));
  _a->out_2_r3 = malloc_check(size_of_test*sizeof(*(_a->out_2_r3)));
  _a->mem_y = malloc_check(size_of_test*sizeof(*(_a->mem_y)));
  _a->mem_x = malloc_check(size_of_test*sizeof(*(_a->mem_x)));
  _a->mem_ok = malloc_check(size_of_test*sizeof(*(_a->mem_ok)));
  if (_a->do_shuffle) {
    _a->_idx = malloc_check(size_of_test*sizeof(*(_a->_idx)));
    for (int _i = _a->_size_of_test-1 ; _i >= 0 ; _i--) {
      _a->_idx[_i] = _i;
    }
  } else {
    for (int _i = _a->_size_of_test-1 ; _i >= 0 ; _i--) {
      _a->y[_i] = &(_a->mem_y[_i]);
      _a->x[_i] = &(_a->mem_x[_i]);
      _a->ok[_i] = &(_a->mem_ok[_i]);
    }
  }
  _a->fst_barrier = pb_create(N);
  _a->s_or = po_create(N);
  for (int _p = N-1 ; _p >= 0 ; _p--) {
    _a->cpy_ok[_p] = malloc_check(size_of_test*sizeof(*(_a->cpy_ok[_p])));
    _a->cpy_y[_p] = malloc_check(size_of_test*sizeof(*(_a->cpy_y[_p])));
  }
  _a->barrier = malloc_check(size_of_test*sizeof(*(_a->barrier)));
}

static void finalize(ctx_t *_a) {
  free((void *)_a->mem_y);
  free((void *)_a->mem_x);
  free((void *)_a->mem_ok);
  if (_a->do_shuffle) {
    free((void *)_a->_idx);
  }
  free((void *)_a->y);
  free((void *)_a->x);
  free((void *)_a->ok);
  free((void *)_a->out_1_r1);
  free((void *)_a->out_2_r3);
  pb_free(_a->fst_barrier);
  po_free(_a->s_or);
  for (int _p = N-1 ; _p >= 0 ; _p--) {
    free((void *)_a->cpy_ok[_p]);
    free((void *)_a->cpy_y[_p]);
  }
  free((void *)_a->barrier);
}

static void reinit(ctx_t *_a) {
  if (_a->do_shuffle) shuffle(_a);
  for (int _i = _a->_size_of_test-1 ; _i >= 0 ; _i--) {
    _a->mem_y[_i] = 0;
    _a->mem_x[_i] = 0;
    _a->mem_ok[_i] = 1;
    _a->out_1_r1[_i] = -239487;
    _a->out_2_r3[_i] = -239487;
    _a->barrier[_i] = 0;
  }
}

/**************************************/
/* Prefetch (and check) global values */
/**************************************/

static void check_globals(ctx_t *_a) {
  int *mem_y = _a->mem_y;
  int *mem_x = _a->mem_x;
  int *mem_ok = _a->mem_ok;
  for (int _i = _a->_size_of_test-1 ; _i >= 0 ; _i--) {
    if (rand_bit(&(_a->seed)) && mem_y[_i] != 0) err(2,"check_globals failed");
    if (rand_bit(&(_a->seed)) && mem_x[_i] != 0) err(2,"check_globals failed");
    if (rand_bit(&(_a->seed)) && mem_ok[_i] != 1) err(2,"check_globals failed");
  }
  pb_wait(_a->fst_barrier);
}


static void stabilize_globals(int _id, ctx_t *_a) {
  int* *ok = _a->ok;
  int **cpy_ok = _a->cpy_ok;
  int* *y = _a->y;
  int **cpy_y = _a->cpy_y;

  pb_wait(_a->fst_barrier); 
  for ( ; ; ) {
    for (int _i = _a->_size_of_test-1 ; _i >= 0 ; _i--) {
      cpy_ok[_id][_i] = *(ok[_i]);
      cpy_y[_id][_i] = *(y[_i]);
    }
    po_reinit(_a->s_or);
    int _found;
    int _nxt_id = (_id+1) % N;
    _found = 0;
    for (int _i = size_of_test-1 ; _i >= 0 && !_found ; _i--) {
      if (cpy_ok[_id][_i] != cpy_ok[_nxt_id][_i]) { _found = 1; break; }
      if (cpy_y[_id][_i] != cpy_y[_nxt_id][_i]) { _found = 1; break; }
    }
    if (_found) { fprintf(stderr,"%i: Stabilizing final state!\n",_id); }
    if (!po_wait(_a->s_or,_found)) return ;
  }
}

/***************/
/* Litmus code */
/***************/

typedef struct {
  int th_id; /* I am running on this thread */
  int *cpu; /* On this cpu */
  ctx_t *_a;   /* In this context */
} parg_t;

static void *P0(void *_vb) {
  mbar();
  parg_t *_b = (parg_t *)_vb;
  ctx_t *_a = _b->_a;
  write_one_affinity(_b->cpu[0]);
  check_globals(_a);
  int _th_id = _b->th_id;
  int volatile *barrier = _a->barrier;
  int* *x = _a->x;
  for (int _i = _a->_size_of_test-1 ; _i >= 0 ; _i--) {
    barrier_wait(_th_id,_i,&barrier[_i]);
    int trashed_r1;
asm __volatile__ (
"\n"
"#START _litmus_P0\n"
"#_litmus_P0_0\n\t"
"li %[r1],1\n"
"#_litmus_P0_1\n\t"
"stw %[r1],0(%[r2])\n"
"#END_litmus\n\t"
:[r1] "=&r" (trashed_r1)
:[r2] "r" (x[_i])
:"cc","memory","r0"
);
  }
  stabilize_globals(0,_a);
  mbar();
  return NULL;
}

static void *P1(void *_vb) {
  mbar();
  parg_t *_b = (parg_t *)_vb;
  ctx_t *_a = _b->_a;
  write_one_affinity(_b->cpu[1]);
  check_globals(_a);
  int _th_id = _b->th_id;
  int volatile *barrier = _a->barrier;
  int* *ok = _a->ok;
  int* *x = _a->x;
  int* *y = _a->y;
  int *out_1_r1 = _a->out_1_r1;
  for (int _i = _a->_size_of_test-1 ; _i >= 0 ; _i--) {
    barrier_wait(_th_id,_i,&barrier[_i]);
    int trashed_r3;
    int trashed_r5;
    int trashed_r7;
asm __volatile__ (
"\n"
"#START _litmus_P1\n"
"#_litmus_P1_0\n\t"
"lwarx %[r1],0,%[r2]\n"
"#_litmus_P1_1\n\t"
"stwcx. %[r1],0,%[r2]\n"
"#_litmus_P1_2\n\t"
"bne Fail1\n"
"#_litmus_P1_3\n\t"
"lwsync\n"
"#_litmus_P1_4\n\t"
"li %[r3],1\n"
"#_litmus_P1_5\n\t"
"lwarx %[r7],0,%[r4]\n"
"#_litmus_P1_6\n\t"
"stwcx. %[r3],0,%[r4]\n"
"#_litmus_P1_7\n\t"
"bne Fail1\n"
"#_litmus_P1_8\n\t"
"b Exit1\n"
"#_litmus_P1_9\n\t"
"Fail1:\n"
"#_litmus_P1_10\n\t"
"li %[r5],0\n"
"#_litmus_P1_11\n\t"
"stw %[r5],0(%[r6])\n"
"#_litmus_P1_12\n\t"
"Exit1:\n"
"#END_litmus\n\t"
:[r1] "=&r" (out_1_r1[_i]),[r7] "=&r" (trashed_r7),[r5] "=&r" (trashed_r5),[r3] "=&r" (trashed_r3)
:[r2] "r" (x[_i]),[r4] "r" (y[_i]),[r6] "r" (ok[_i])
:"cc","memory","r0"
);
  }
  stabilize_globals(1,_a);
  mbar();
  return NULL;
}

static void *P2(void *_vb) {
  mbar();
  parg_t *_b = (parg_t *)_vb;
  ctx_t *_a = _b->_a;
  write_one_affinity(_b->cpu[2]);
  check_globals(_a);
  int _th_id = _b->th_id;
  int volatile *barrier = _a->barrier;
  int* *ok = _a->ok;
  int* *x = _a->x;
  int* *y = _a->y;
  int *out_2_r3 = _a->out_2_r3;
  for (int _i = _a->_size_of_test-1 ; _i >= 0 ; _i--) {
    barrier_wait(_th_id,_i,&barrier[_i]);
    int trashed_r1;
    int trashed_r5;
asm __volatile__ (
"\n"
"#START _litmus_P2\n"
"#_litmus_P2_0\n\t"
"li %[r1],2\n"
"#_litmus_P2_1\n\t"
"stw %[r1],0(%[r2])\n"
"#_litmus_P2_2\n\t"
"sync\n"
"#_litmus_P2_3\n\t"
"lwarx %[r3],0,%[r4]\n"
"#_litmus_P2_4\n\t"
"stwcx. %[r3],0,%[r4]\n"
"#_litmus_P2_5\n\t"
"bne Fail2\n"
"#_litmus_P2_6\n\t"
"b Exit2\n"
"#_litmus_P2_7\n\t"
"Fail2:\n"
"#_litmus_P2_8\n\t"
"li %[r5],0\n"
"#_litmus_P2_9\n\t"
"stw %[r5],0(%[r6])\n"
"#_litmus_P2_10\n\t"
"Exit2:\n"
"#END_litmus\n\t"
:[r3] "=&r" (out_2_r3[_i]),[r5] "=&r" (trashed_r5),[r1] "=&r" (trashed_r1)
:[r2] "r" (y[_i]),[r4] "r" (x[_i]),[r6] "r" (ok[_i])
:"cc","memory","r0"
);
  }
  stabilize_globals(2,_a);
  mbar();
  return NULL;
}

typedef struct {
  int max_run;
  pm_t *p_mutex;
  pb_t *p_barrier;
  int aff_rand;
  int z_id;
  int ncpus;
  int ncpus_used;
  int *cpus;
  int do_shuffle;
  int do_change;
  cpus_t *cm;
} zyva_t;

#define NT N

static void *zyva(void *_va) {
  zyva_t *_a = (zyva_t *) _va;
  int max_run = _a->max_run;
  pb_wait(_a->p_barrier);
  pthread_t thread[NT];
  parg_t parg[N];
  f_t *fun[] = {&P0,&P1,&P2};
  hist_t *hist = alloc_hist();
  ctx_t  ctx;
  ctx.do_shuffle = _a->do_shuffle;

  init(&ctx);
  ctx.cm = _a->cm;
  for (int _p = N-1 ; _p >= 0 ; _p--) {
    parg[_p].th_id = _p; parg[_p]._a = &ctx;
    parg[_p].cpu = &(_a->cpus[0]);
  }

  for (int n_run = 0 ; n_run < max_run ; n_run++) {
    if (_a->aff_rand) {
      pb_wait(_a->p_barrier);
      if (stop_now) {
        break;
      }
      if (_a->z_id == 0) perm_prefix_ints(&ctx.seed,_a->cpus,_a->ncpus_used,_a->ncpus);
      pb_wait(_a->p_barrier);
    } else {
      if (stop_now) {
        break;
      }
    }
    if (verbose>1) fprintf(stderr, "Run %i of %i\r", n_run, max_run);
    reinit(&ctx);
    if (_a->do_change) perm_funs(&ctx.seed,fun,N);
    for (int _p = NT-1 ; _p >= 0 ; _p--) {
      launch(&thread[_p],fun[_p],&parg[_p]);
    }
    if (_a->do_change) perm_threads(&ctx.seed,thread,NT);
    for (int _p = NT-1 ; _p >= 0 ; _p--) {
      join(&thread[_p]);
    }
    /* Log final states */
    for (int _i = ctx._size_of_test-1 ; _i >= 0 ; _i--) {
      int _out_1_r1_i = ctx.out_1_r1[_i];
      int _out_2_r3_i = ctx.out_2_r3[_i];
      int _ok_i = *(ctx.ok[_i]);
      int _y_i = *(ctx.y[_i]);
      outcome_t o;
      int cond;

      for (int _p = N-1 ; _p >= 0 ; _p--) {
        if (_ok_i != ctx.cpy_ok[_p][_i]) err(1,"global ok unstabilized") ;
      }
      for (int _p = N-1 ; _p >= 0 ; _p--) {
        if (_y_i != ctx.cpy_y[_p][_i]) err(1,"global y unstabilized") ;
      }
      o[out_1_r1_f] = _out_1_r1_i;
      o[out_2_r3_f] = _out_2_r3_i;
      o[ok_f] = _ok_i;
      o[y_f] = _y_i;
      add_outcome(hist,1,o);
      cond = (((_ok_i == 1) && (_y_i == 2)) && (_out_1_r1_i == 1)) && (_out_2_r3_i == 0);
      if (cond) { hist->n_pos++; } else { hist->n_neg++; }
    }
    if (speedcheck && hist->n_pos > 0) {
      stop_now = 1;
    }
  }

  finalize(&ctx);
  return hist;
}

static int color_0[] = {0, 1, -1};
static int color_1[] = {2, -1};
static int *color[] = {color_0, color_1, NULL};
static int diff[] = {2, 0, -1};

int main(int argc, char **argv) {
  cpus_t *def_all_cpus = read_affinity();
  tsc_t start = timeofday();
  cmd_t def = { 0, NUMBER_OF_RUN, SIZE_OF_TEST, -1, AVAIL, 0, SYNC_N, AFF_INCR, def_all_cpus, 0, 0, 1, 0, NULL, NULL, -1, 1};
  cmd_t cmd = def;

  parse_cmd(argc,argv,&def,&cmd);
  verbose = cmd.verbose;
  size_of_test = cmd.size_of_test;
  speedcheck = cmd.speedcheck;
  int n_avail = cmd.avail > 0 ? cmd.avail : cmd.aff_cpus->sz;
  if (n_avail >  cmd.aff_cpus->sz) fprintf(stderr,"Warning: avail=%i, available=%i\n",n_avail, cmd.aff_cpus->sz);
  int n_exe;
  if (cmd.n_exe > 0) {
    n_exe = cmd.n_exe;
  } else {
    n_exe = n_avail < N ? 1 : n_avail / N;
  }
  cpus_t *cm = coremap_seq(def_all_cpus->sz,2);
  if (verbose) {
    fprintf(stderr, "WRW+WR+lwsyncaa+syncpa: n=%i, r=%i, s=%i",n_exe,cmd.max_run,size_of_test);
    fprintf(stderr,", %crm",cmd.shuffle?'+':'-');
    fprintf(stderr, ", i=%i, %sra",cmd.aff_incr,cmd.aff_rand ? "+" : "-");
    fprintf(stderr, ", %sca",cmd.aff_custom ? "+" : "-");
    fprintf(stderr,", p='");
    cpus_dump(stderr,cmd.aff_cpus);
    fprintf(stderr,"'");
    fprintf(stderr,"\n");
    if (verbose > 1 && cm) {
      fprintf(stderr,"logical proc -> core: ");
      cpus_dump(stderr,cm);
      fprintf(stderr,"\n");
    }
  }
  
  cpus_t *all_cpus = cmd.aff_cpus;
  int aff_cpus_sz = cmd.aff_rand ? max(all_cpus->sz,N*n_exe) : N*n_exe;
  int aff_cpus[aff_cpus_sz];
  if (cmd.aff_rand) {
    for (int k = 0 ; k < aff_cpus_sz ; k++) {
      aff_cpus[k] = all_cpus->cpu[k % all_cpus->sz];
    }
  } else if (cmd.aff_custom) {
    custom_affinity(cm,color,diff,all_cpus,n_exe);
    if (verbose) {
      fprintf(stderr,"thread allocation: \n");
      cpus_dump_test(stderr,all_cpus,cm,N);
    }
  }
  pthread_t th[n_exe];
  zyva_t zarg[n_exe];
  pm_t *p_mutex = pm_create();
  pb_t *p_barrier = pb_create(n_exe);
  int next_cpu = 0;
  int delta = cmd.aff_custom ? 1 : cmd.aff_incr;
  if (delta <= 0) {
    for (int k=0 ; k < all_cpus->sz ; k++) all_cpus->cpu[k] = -1;
    delta = 1;
  } else {
    delta %= all_cpus->sz;
  }
  int start_scan=0, max_start=gcd(delta,all_cpus->sz);
  int *aff_p = aff_cpus;
  for (int k=0 ; k < n_exe ; k++) {
    zyva_t *p = &zarg[k];
    p->max_run = cmd.max_run; p->p_mutex = p_mutex; p->p_barrier = p_barrier; 
    p->do_shuffle = cmd.shuffle;
    p->do_change = !(cmd.aff_custom || cmd.aff_rand);
    p->cm = cm;
    p->z_id = k;
    p->aff_rand = cmd.aff_rand;
    p->ncpus = aff_cpus_sz;
    p->ncpus_used = N*n_exe;
    p->cpus = aff_p;
    if (cmd.aff_rand) {
      aff_p += N;
    } else {
      for (int i=0 ; i < N ; i++) {
        *aff_p = all_cpus->cpu[next_cpu]; aff_p++;
        next_cpu += delta; next_cpu %= all_cpus->sz;
        if (next_cpu == start_scan) {
          start_scan++ ; start_scan %= max_start;
          next_cpu = start_scan;
        }
      }
    }
    launch(&th[k],zyva,p);
  }
  cpus_free(all_cpus);

  count_t n_outs = size_of_test; n_outs *= cmd.max_run;
  hist_t *hist = (hist_t *)join(&th[0]);
  for (int k=1 ; k < n_exe ; k++) {
    hist_t *hk = (hist_t *)join(&th[k]);
    if (!speedcheck) {
      if (sum_hist(hk) != n_outs || hk->n_pos + hk->n_neg != n_outs) {
        err(1,"sum_hist");
      }
    }
    merge_hists(hist,hk);
    free_hist(hk);
  }
  tsc_t total = timeofday() - start ;
  pm_free(p_mutex);
  pb_free(p_barrier);

  n_outs *= n_exe ;
  if (!speedcheck) {
    if (sum_hist(hist) != n_outs || hist->n_pos + hist->n_neg != n_outs) {
      err(1,"sum_hist") ;
    }
  }
  int cond = dump_hist(stdout,hist);
  count_t p_true = hist->n_pos, p_false = hist->n_neg ;
  free_hist(hist);
  cpus_free(cm);
  fprintf(stdout,"%s=%s\n","Hash","9b468fae5694038351741c4f7f6afedb");
  fprintf(stdout,"%s=%s\n","Cycle","RfePA LwSyncdRWAA WseAP SyncdWRPA FreAP");
  fprintf(stdout,"%s=%s\n","Com","Rf Ws Fr");
  fprintf(stdout,"%s=%s\n","Orig","RfePA LwSyncdRWAA WseAP SyncdWRPA FreAP");
  if (cmd.aff_custom) {
    fprintf(stdout,"%s=%s\n","Affinity","[0, 1] [2] ; (2,0)");
  }
  fprintf(stdout,"Observation WRW+WR+lwsyncaa+syncpa %s %"PCTR" %"PCTR"\n",!p_true ? "Never" : !p_false ? "Always" : "Sometimes",p_true,p_false) ;
  fprintf(stdout,"Time WRW+WR+lwsyncaa+syncpa %.2f\n",total / 1000000.0) ;
  exit(cond ^ cond) ;
}
